#include "wiced.h"
#include "command_console.h"
#include "command_console_ping.h"
#include "command_console_platform.h"
#include "command_console_wifi.h"
#include "command_console_dct.h"
#include "console_wl.h"

/******************************************************
 *                      Macros
 ******************************************************/

#define DIAGNOSTICS_COMMANDS \
    {"start",          start_tx,       0, NULL, NULL, "[-c ip_address] [-p port] [-n packet_count]", "start transmission"}, \


/******************************************************
 *                    Constants
 ******************************************************/

#define MAX_LINE_LENGTH                        128
#define MAX_HISTORY_LENGTH                     20
#define ARG_BUFFER_LEN                         31
#define TXDATATASK_STACK_SIZE                  4096
#define TXDATATASK_PRIORITY                    WICED_APPLICATION_PRIORITY
#define PWM_FREQ_HZ                            66000
#define PWM_H_PERIOD_SEC                       4.0e-6
#define PWM_DUTY_CYCLE                         (float)(((PWM_H_PERIOD_SEC) * (PWM_FREQ_HZ)) * 100.0)
#define TCP_CONNECTION_NUMBER_OF_RETRIES       8
#define TCP_CLIENT_CONNECT_TIMEOUT             1000
#define TRUE                                   1
#define FALSE                                  0
#define DEFAULT_STARTCMD_TCP_SERVER_IP_ADDRESS MAKE_IPV4_ADDRESS(192,168,3,100)
#define DEFAULT_STARTCMD_TCP_SERVER_PORT       5001
#define DEFAULT_STARTCMD_TX_PACKET_COUNT       100
#define DATA_CNT                               (4096*3)
#define INTR_INTERVAL_TH                       20000 // threshold 20us

/******************************************************
 *                 Type Definitions
 ******************************************************/

typedef unsigned char  Bool;

typedef struct StartCmd
{
    int32_t          numberOfPackets;       /* Number of packets to send */
    uint8_t          *ip;                   /* IP address */
    uint16_t         portNumber;            /* Server's port address*/
    uint8_t          ipv6;                  /* IPV4 or IPv6 flag. By default IPV4 enable */
} StartCmd_t;

typedef struct Rec {
    int buf_no;
} Rec;

/******************************************************
 *                    Externs
 ******************************************************/

extern const platform_gpio_t       platform_gpio_pins[];

/******************************************************
 *               Static Variable Definitions
 ******************************************************/

static char                line_buffer[MAX_LINE_LENGTH];
static char                history_buffer_storage[MAX_LINE_LENGTH * MAX_HISTORY_LENGTH];
static wiced_tcp_socket_t  StartCmd_sock;

/******************************************************
 *               Variable Definitions
 ******************************************************/

Bool              is_socket_opened = FALSE;
StartCmd_t        start_cmd_params;
wiced_thread_t    tx_data_task_handle;
wiced_queue_t     send_buf_queue_handle;
Rec               send_buf_queue_data;
int16_t           data_buf[DATA_CNT];
int               count_device_int = 0;
int               wlan_send_countdown = 0;
uint32_t          cur_interval;
int               exceed_cnt = 0;
int               prev_exceed_cnt = 0;
uint32_t          exceed_val = 0;
uint64_t          prev_cpu_time = 0;


void intrCallback(void *arg)
{
    int      is_exceed;
    uint64_t cur_cpu_time;

    count_device_int++;

    wiced_gpio_output_high(WICED_GPIO_1);   // WICED_GPIO_1: GPIO_0:  J10.1 HI
    cur_cpu_time = platform_get_nanosecond_clock_value();

    cur_interval = (uint32_t)(cur_cpu_time - prev_cpu_time);
    is_exceed = ((prev_cpu_time > 0) && (cur_interval > INTR_INTERVAL_TH)) ? TRUE : FALSE;
    if (is_exceed) {
        exceed_val = cur_interval;
        exceed_cnt++;
    }
    prev_cpu_time = cur_cpu_time;

    // TX
    if (count_device_int >= DATA_CNT) {
        wiced_rtos_push_to_queue(&send_buf_queue_handle, &send_buf_queue_data, WICED_NO_WAIT);
        count_device_int = 0;
    }

    wiced_gpio_output_low(WICED_GPIO_1);    // WICED_GPIO_1: GPIO_0:  J10.1 LO
}

static void txDataTask(wiced_thread_arg_t arg)
{
    for (;;) {
            Rec tmp_msg;
            wiced_rtos_pop_from_queue(&send_buf_queue_handle, &tmp_msg, WICED_WAIT_FOREVER);

            if (wlan_send_countdown) {
                wlan_send_countdown--;

                if (exceed_cnt > prev_exceed_cnt) {
                    WPRINT_APP_INFO(("Unexpected longer interrupt interval detected at wlan_send_countdown=%d, exceed_cnt=%d, interval=%luns\n", wlan_send_countdown, exceed_cnt, exceed_val));
                }
                prev_exceed_cnt = exceed_cnt;
                // send (blocking)
                wiced_tcp_send_buffer(&StartCmd_sock, &data_buf[0], DATA_CNT * sizeof(uint16_t));
                // del socket
                if ((wlan_send_countdown == 0) && is_socket_opened) {
                    wiced_tcp_delete_socket(&StartCmd_sock);
                    is_socket_opened = FALSE;
                    WPRINT_APP_INFO(("Transmission finished\n"));
                }
            }
            wiced_gpio_output_low(WICED_GPIO_2);    // WICED_GPIO_2: GPIO_1:  J10.2 LO
    }
}

/*
 * start -c <ip_address> [-p <port_number>] [-n <number_of_tx_packets>]
 */
int start_tx(int argc, char* argv[])
{
    wiced_result_t     result;
    wiced_ip_address_t INITIALISER_IPV4_ADDRESS(tx_target_ip_address, DEFAULT_STARTCMD_TCP_SERVER_IP_ADDRESS);
    int                temp[4];
    int                connection_try_count;

    if (wlan_send_countdown != 0) {
        WPRINT_APP_INFO(("Busy\n"));
        return WICED_SUCCESS;
    }
    start_cmd_params.portNumber = DEFAULT_STARTCMD_TCP_SERVER_PORT;
    start_cmd_params.numberOfPackets = DEFAULT_STARTCMD_TX_PACKET_COUNT;
    for (int i = 1; i < argc; i++) {
        switch (argv[i][1]) {
            case 'c':
                sscanf(argv[i+1], "%d.%d.%d.%d", &temp[0], &temp[1], &temp[2], &temp[3]);
                tx_target_ip_address.ip.v4 = temp[0] << 24 | temp[1] << 16 | temp[2] << 8 | temp[3];
                i++;
                break;

            case 'p':
                start_cmd_params.portNumber = atoi(argv[i+1]);
                i++;
                break;

            case 'n':
                start_cmd_params.numberOfPackets = atoi(argv[i+1]);
                i++;
                break;

            default:
                WPRINT_APP_INFO(("Not supported, ignoring: %s\n", argv[i]));
            break;
        }
    }

    /* Create a TCP socket */
    wiced_tcp_create_socket(&StartCmd_sock, WICED_STA_INTERFACE);
    is_socket_opened = TRUE;

    /* Bind to the socket */
    wiced_tcp_bind(&StartCmd_sock, start_cmd_params.portNumber);

    /* Connect to the remote TCP server */
    WPRINT_APP_INFO(("wiced_tcp_connect start\n"));
    connection_try_count = 0;
    do {
        result = wiced_tcp_connect(&StartCmd_sock, &tx_target_ip_address, start_cmd_params.portNumber, TCP_CLIENT_CONNECT_TIMEOUT);
        connection_try_count++;
    } while ((result != WICED_SUCCESS) && (connection_try_count < TCP_CONNECTION_NUMBER_OF_RETRIES));
    if (result != WICED_SUCCESS) {
        WPRINT_APP_INFO(("wiced_tcp_connect ERROR result=%d\n\r", result));
        wiced_tcp_delete_socket(&StartCmd_sock);
    } else {
        wlan_send_countdown = start_cmd_params.numberOfPackets;
    }

    return WICED_SUCCESS;
}

static const command_t init_commands[] = {
    DIAGNOSTICS_COMMANDS
    PLATFORM_COMMANDS
    WIFI_COMMANDS
    DCT_CONSOLE_COMMANDS
    WL_COMMANDS
    CMD_TABLE_END
};

void application_start()
{
    wiced_init();
    platform_init_nanosecond_clock();
    wiced_gpio_init(WICED_GPIO_6, INPUT_PULL_DOWN);   // WICED_GPIO_6:  GPIO_10: J12.4
    wiced_gpio_init(WICED_GPIO_1, OUTPUT_PUSH_PULL);  // WICED_GPIO_1:  GPIO_0:  J10.1
    wiced_gpio_init(WICED_GPIO_10,OUTPUT_PUSH_PULL);  // WICED_GPIO_10: GPIO_14: J10.5
    wiced_gpio_init(WICED_GPIO_2, OUTPUT_PUSH_PULL);  // WICED_GPIO_2:  GPIO_1:  J10.2
    wiced_gpio_output_low(WICED_GPIO_1);    // WICED_GPIO_1:  GPIO_0:  J10.1 LO
    wiced_gpio_output_low(WICED_GPIO_10);   // WICED_GPIO_10: GPIO_14: J10.5 LO
    wiced_gpio_output_low(WICED_GPIO_2);    // WICED_GPIO_2:  GPIO_1:  J10.2 LO

    wiced_rtos_init_queue(&send_buf_queue_handle, "sendBufQueue", sizeof(Rec), 1);
    send_buf_queue_data.buf_no = 0;

    wiced_rtos_create_thread(&tx_data_task_handle, TXDATATASK_PRIORITY, "txDataTask", txDataTask, TXDATATASK_STACK_SIZE, NULL);

    wiced_gpio_input_irq_enable(WICED_GPIO_6, IRQ_TRIGGER_RISING_EDGE, intrCallback, NULL);
    wiced_pwm_init(WICED_PWM_4, PWM_FREQ_HZ, PWM_DUTY_CYCLE);  // WICED_PWM_4(GPIO_11) J10.7
    wiced_pwm_start(WICED_PWM_4);

    command_console_init(STDIO_UART, MAX_LINE_LENGTH, line_buffer, MAX_HISTORY_LENGTH, history_buffer_storage, " ");
    console_add_cmd_table(init_commands);
}
